from collections.abc import Callable, Sequence
from typing import Any

import aioinject
from aioinject import Provider
from aioinject.validation.abc import ContainerValidator
from aioinject.validation.error import (
    ContainerValidationError,
    DependencyNotFoundError,
)


def all_dependencies_are_present(
    container: aioinject.Container,
) -> Sequence[ContainerValidationError]:
    errors = []
    for providers in container.providers.values():
        for provider in providers:
            for dependency in provider.dependencies:
                if dependency.type_ not in container.providers:
                    error = DependencyNotFoundError(
                        message=f"Provider for type {dependency.type_} not found",
                        dependency=dependency.type_,
                    )
                    errors.append(error)

    return errors


class ForbidDependency(ContainerValidator):
    def __init__(
        self,
        dependant: Callable[[Provider[Any]], bool],
        dependency: Callable[[Provider[Any]], bool],
    ) -> None:
        self.dependant = dependant
        self.dependency = dependency

    def __call__(
        self,
        container: aioinject.Container,
    ) -> Sequence[ContainerValidationError]:
        errors = []
        for providers in container.providers.values():
            for provider in providers:
                if not self.dependant(provider):
                    continue

                for dependency in provider.dependencies:
                    dependency_provider = container.get_provider(
                        type_=dependency.type_,
                        impl=dependency.implementation,
                    )
                    if self.dependency(dependency_provider):
                        msg = f"Provider {provider!r} cannot depend on {dependency_provider!r}"
                        errors.append(ContainerValidationError(msg))
        return errors



from __future__ import annotations

import contextlib
import contextvars
import inspect
from collections.abc import Callable, Coroutine, Iterable
from contextlib import AsyncExitStack, ExitStack
from contextvars import ContextVar
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    TypeAlias,
    TypeVar,
    Union,
    overload,
)

from typing_extensions import Self

from aioinject.providers import Dependency
from aioinject.utils import enter_context_maybe


if TYPE_CHECKING:
    from aioinject.containers import Container

_T = TypeVar("_T")

_AnyCtx: TypeAlias = Union["InjectionContext", "SyncInjectionContext"]
_TypeAndImpl: TypeAlias = tuple[type[_T], _T | None]
_ExitStackT = TypeVar("_ExitStackT")

context_var: ContextVar[_AnyCtx] = ContextVar("aioinject_context")
container_var: ContextVar[Container] = ContextVar("aioinject_container")


class _BaseInjectionContext(Generic[_ExitStackT]):
    _token: contextvars.Token[_AnyCtx] | None
    _exit_stack_type: type[_ExitStackT]

    def __init__(self, container: Container) -> None:
        self._container = container
        self._exit_stack = self._exit_stack_type()
        self._cache: dict[_TypeAndImpl[Any], Any] = {}
        self._token = None

    def __class_getitem__(
        cls,
        item: type[_ExitStackT],
    ) -> _BaseInjectionContext[type[_ExitStackT]]:
        return type(  # type: ignore[return-value]
            f"_BaseInjectionContext[{item.__class__.__name__}]",
            (cls,),
            {"_exit_stack_type": item},
        )


class InjectionContext(_BaseInjectionContext[AsyncExitStack]):
    async def resolve(
        self,
        type_: type[_T],
        impl: Any | None = None,
        *,
        use_cache: bool = True,
    ) -> _T:
        if use_cache and (type_, impl) in self._cache:
            return self._cache[type_, impl]

        provider = self._container.get_provider(type_, impl)
        dependencies = {
            dep.name: await self.resolve(
                type_=dep.type_,
                impl=dep.implementation,
                use_cache=dep.use_cache,
            )
            for dep in provider.dependencies
        }

        resolved = await enter_context_maybe(
            resolved=await provider.provide(**dependencies),
            stack=self._exit_stack,
        )
        if use_cache:
            self._cache[type_, impl] = resolved
        return resolved

    @overload
    async def execute(
        self,
        function: Callable[..., Coroutine[Any, Any, _T]],
        dependencies: Iterable[Dependency[object]],
        *args: Any,
        **kwargs: Any,
    ) -> _T:
        ...

    @overload
    async def execute(
        self,
        function: Callable[..., _T],
        dependencies: Iterable[Dependency[object]],
        *args: Any,
        **kwargs: Any,
    ) -> _T:
        ...

    async def execute(
        self,
        function: Callable[..., Coroutine[Any, Any, _T] | _T],
        dependencies: Iterable[Dependency[object]],
        *args: Any,
        **kwargs: Any,
    ) -> _T:
        resolved = {}
        for dependency in dependencies:
            if dependency.name in kwargs:
                continue
            resolved[dependency.name] = await self.resolve(
                type_=dependency.type_,
                impl=dependency.implementation,
                use_cache=dependency.use_cache,
            )
        if inspect.iscoroutinefunction(function):
            return await function(*args, **kwargs, **resolved)
        return function(*args, **kwargs, **resolved)  # type: ignore[return-value]

    async def __aenter__(self) -> Self:
        self._token = context_var.set(self)
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        context_var.reset(self._token)  # type: ignore[arg-type]
        await self._exit_stack.__aexit__(exc_type, exc_val, exc_tb)


class SyncInjectionContext(_BaseInjectionContext[ExitStack]):
    def resolve(
        self,
        type_: type[_T],
        impl: Any | None = None,
        *,
        use_cache: bool = True,
    ) -> _T:
        if use_cache and (type_, impl) in self._cache:
            return self._cache[type_, impl]

        provider = self._container.get_provider(type_, impl)
        dependencies = {
            dep.name: self.resolve(
                type_=dep.type_,
                impl=dep.implementation,
                use_cache=dep.use_cache,
            )
            for dep in provider.dependencies
        }

        resolved = provider.provide_sync(**dependencies)
        if isinstance(resolved, contextlib.ContextDecorator):
            resolved = self._exit_stack.enter_context(resolved)  # type: ignore[arg-type]
        if use_cache:
            self._cache[type_, impl] = resolved
        return resolved

    def execute(
        self,
        function: Callable[..., _T],
        dependencies: Iterable[Dependency[object]],
        *args: Any,
        **kwargs: Any,
    ) -> _T:
        resolved = {}
        for dependency in dependencies:
            if dependency.name in kwargs:
                continue
            resolved[dependency.name] = self.resolve(
                type_=dependency.type_,
                impl=dependency.implementation,
                use_cache=dependency.use_cache,
            )
        return function(*args, **kwargs, **resolved)

    def __enter__(self) -> Self:
        self._token = context_var.set(self)
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        context_var.reset(self._token)  # type: ignore[arg-type]
        self._exit_stack.__exit__(exc_type, exc_val, exc_tb)



import contextlib
from collections.abc import AsyncIterator

import pytest

from aioinject import Singleton, providers
from aioinject.containers import Container
from aioinject.context import InjectionContext


class _AbstractService:
    pass


class _ServiceA(_AbstractService):
    pass


class _ServiceB(_AbstractService):
    pass


@pytest.fixture
def container() -> Container:
    return Container()


def test_can_init(container: Container) -> None:
    assert container


def test_can_retrieve_context(container: Container) -> None:
    ctx = container.context()
    assert isinstance(ctx, InjectionContext)


def test_can_register_single(container: Container) -> None:
    provider = providers.Scoped(_ServiceA)
    container.register(provider)

    expected = {_ServiceA: [provider]}
    assert container.providers == expected


def test_can_register_multi(container: Container) -> None:
    provider_a = providers.Scoped(_ServiceA)
    provider_b = providers.Scoped(_ServiceB)
    container.register(provider_a)
    container.register(provider_b)

    expected = {_ServiceA: [provider_a], _ServiceB: [provider_b]}
    assert container.providers == expected


def test_can_retrieve_single_provider(container: Container) -> None:
    int_provider = providers.Scoped(int)
    container.register(int_provider)
    assert container.get_provider(int)


@pytest.fixture
def multi_provider_container(container: Container) -> Container:
    a_provider = providers.Scoped(_ServiceA, type_=_AbstractService)
    b_provider = providers.Scoped(_ServiceB, type_=_AbstractService)
    container.register(a_provider)
    container.register(b_provider)
    return container


def test_get_provider_raises_error_if_multiple_providers(
    multi_provider_container: Container,
) -> None:
    with pytest.raises(ValueError) as exc_info:  # noqa: PT011
        assert multi_provider_container.get_provider(_AbstractService)

    msg = (
        f"Multiple providers for type {_AbstractService.__qualname__} were found, "
        f"you have to specify implementation using 'impl' parameter: "
        f"Annotated[IService, Inject(impl=Service)]"
    )
    assert str(exc_info.value) == msg


def test_can_get_multi_provider_if__specified(
    multi_provider_container: Container,
) -> None:
    assert multi_provider_container.get_provider(_AbstractService, _ServiceA)
    assert multi_provider_container.get_provider(_AbstractService, _ServiceB)


def test_missing_provider() -> None:
    container = Container()
    with pytest.raises(ValueError) as exc_info:  # noqa: PT011
        assert container.get_provider(_ServiceA)

    msg = f"Provider for type {_ServiceA.__qualname__} not found"
    assert str(exc_info.value) == msg


@pytest.mark.anyio
async def test_should_close_singletons() -> None:
    shutdown = False

    @contextlib.asynccontextmanager
    async def dependency() -> AsyncIterator[int]:
        nonlocal shutdown

        yield 42
        shutdown = True

    provider = Singleton(dependency)

    container = Container()
    container.register(provider)
    for _ in range(2):
        async with container.context() as ctx:
            assert await ctx.resolve(int) == 42  # noqa: PLR2004
    assert shutdown is False

    await container.aclose()
    assert shutdown is True



import contextlib
from collections import defaultdict
from collections.abc import Iterator
from typing import Any, TypeAlias, TypeVar

from aioinject.context import InjectionContext, SyncInjectionContext
from aioinject.providers import Provider, Singleton


_T = TypeVar("_T")
_Providers: TypeAlias = dict[type[_T], list[Provider[_T]]]


class Container:
    def __init__(self) -> None:
        self.providers: _Providers[Any] = defaultdict(list)
        self._overrides: _Providers[Any] = defaultdict(list)

    def register(
        self,
        provider: Provider[Any],
    ) -> None:
        self.providers[provider.type_].append(provider)

    @staticmethod
    def _get_provider(
        providers: _Providers[_T],
        type_: type[_T],
        impl: Any | None = None,
    ) -> Provider[_T] | None:
        type_providers = providers[type_]
        if not type_providers:
            return None

        if impl is None and len(type_providers) == 1:
            return type_providers[0]

        if impl is None:
            err_msg = (
                f"Multiple providers for type {type_.__qualname__} were found, "
                f"you have to specify implementation using 'impl' parameter: "
                f"Annotated[IService, Inject(impl=Service)]"
            )
            raise ValueError(err_msg)
        return next((p for p in type_providers if p.impl is impl), None)

    def get_provider(
        self,
        type_: type[_T],
        impl: Any | None = None,
    ) -> Provider[_T]:
        overridden_provider = self._get_provider(self._overrides, type_, impl)
        if overridden_provider is not None:
            return overridden_provider

        provider = self._get_provider(self.providers, type_, impl)
        if provider is None:
            err_msg = f"Provider for type {type_.__qualname__} not found"
            raise ValueError(err_msg)
        return provider

    def context(self) -> InjectionContext:
        return InjectionContext(container=self)

    def sync_context(self) -> SyncInjectionContext:
        return SyncInjectionContext(container=self)

    @contextlib.contextmanager
    def override(
        self,
        provider: Provider[Any],
    ) -> Iterator[None]:
        self._overrides[provider.type_].append(provider)
        yield
        self._overrides[provider.type_].remove(provider)

    async def aclose(self) -> None:
        for providers in self.providers.values():
            for provider in providers:
                if isinstance(provider, Singleton):
                    await provider.aclose()



import uuid

import httpx
import pytest

import aioinject


@pytest.mark.anyio
async def test_function_route(
    http_client: httpx.AsyncClient,
    provided_value: int,
) -> None:
    response = await http_client.get("/function-route")
    assert response.status_code == httpx.codes.OK.value
    assert response.json() == {"value": provided_value}


@pytest.mark.anyio
async def test_function_route_override(
    http_client: httpx.AsyncClient,
    container: aioinject.Container,
) -> None:
    expected = str(uuid.uuid4())
    with container.override(aioinject.Object(expected, type_=int)):
        response = await http_client.get("/function-route")
    assert response.status_code == httpx.codes.OK.value
    assert response.json() == {"value": expected}



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, ParamSpec, TypeVar

from fastapi import Request
from starlette.middleware.base import (
    BaseHTTPMiddleware,
    DispatchFunction,
    RequestResponseEndpoint,
)
from starlette.responses import Response
from starlette.types import ASGIApp

from aioinject import decorators, utils


if TYPE_CHECKING:
    from aioinject.containers import Container

_T = TypeVar("_T")
_P = ParamSpec("_P")


def inject(function: Callable[_P, _T]) -> Callable[_P, _T]:
    wrapper = decorators.inject(
        function,
        inject_method=decorators.InjectMethod.context,
    )
    return utils.clear_wrapper(wrapper)


class AioInjectMiddleware(BaseHTTPMiddleware):
    def __init__(
        self,
        app: ASGIApp,
        container: Container,
        dispatch: DispatchFunction | None = None,
    ) -> None:
        super().__init__(app, dispatch)
        self.container = container

    async def dispatch(
        self,
        request: Request,
        call_next: RequestResponseEndpoint,
    ) -> Response:
        async with self.container.context():
            return await call_next(request)



from typing import Annotated

from aioinject import Inject
from aioinject.utils import get_inject_annotations


def test_inject_annotations_returns_all_inject_markers() -> None:
    def func(
        a: int,  # noqa: ARG001
        b: Annotated[int, Inject],  # noqa: ARG001
        c: Annotated[int, Inject(impl=bool)],  # noqa: ARG001
    ) -> None:
        pass

    assert get_inject_annotations(func) == {
        "b": Annotated[int, Inject],
        "c": Annotated[int, Inject(impl=bool)],
    }



from collections.abc import AsyncIterator
from typing import Annotated

import httpx
import pytest
from litestar import Litestar, get

import aioinject
from aioinject import Inject
from aioinject.ext.litestar import AioInjectPlugin, inject


@pytest.fixture
def app(container: aioinject.Container) -> Litestar:
    @get("/function-route")
    @inject
    async def function_route(
        provided: Annotated[int, Inject],
    ) -> dict[str, str | int]:
        return {"value": provided}

    return Litestar(
        [function_route],
        plugins=[AioInjectPlugin(container=container)],
    )


@pytest.fixture
async def http_client(app: Litestar) -> AsyncIterator[httpx.AsyncClient]:
    async with httpx.AsyncClient(app=app, base_url="http://test") as client:
        yield client



from typing import Annotated

import pytest

from aioinject import Container, Inject, inject, providers
from aioinject.context import container_var
from aioinject.decorators import InjectMethod


class _Session:
    pass


class _Service:
    def __init__(self, session: Annotated[_Session, Inject]) -> None:
        self.session = session


class _Interface:
    def __init__(self, session: Annotated[_Session, Inject]) -> None:
        self.session = session


class _ImplementationA(_Interface):
    pass


class _ImplementationB(_Interface):
    pass


def _get_impl_b(session: Annotated[_Session, Inject]) -> _ImplementationB:
    return _ImplementationB(session)


class _NeedsMultipleImplementations:
    def __init__(
        self,
        a: Annotated[_Interface, Inject(_ImplementationA)],
        b: Annotated[_Interface, Inject(_get_impl_b)],
    ) -> None:
        self.a = a
        self.b = b


@pytest.fixture
def container() -> Container:
    container = Container()
    container.register(providers.Scoped(_Session))
    container.register(providers.Scoped(_Service))
    container.register(providers.Scoped(_ImplementationA, type_=_Interface))
    container.register(providers.Scoped(_get_impl_b, type_=_Interface))
    container.register(providers.Scoped(_NeedsMultipleImplementations))
    return container


@inject
def _injectee(
    test: Annotated[_Session, Inject],
    test_no_cache: Annotated[_Session, Inject(cache=False)],
) -> tuple[_Session, _Session]:
    return test, test_no_cache


def test_would_fail_without_active_context() -> None:
    with pytest.raises(LookupError):
        _injectee()  # type: ignore[call-arg]


def test_would_not_inject_without_inject_marker(container: Container) -> None:
    @inject
    def injectee(session: _Session) -> _Session:
        return session

    with container.sync_context(), pytest.raises(TypeError):
        injectee()  # type: ignore[call-arg]


def test_simple_inject(container: Container) -> None:
    with container.sync_context():
        session, *_ = _injectee()  # type: ignore[call-arg]
        assert isinstance(session, _Session)


def test_no_cache_marker(container: Container) -> None:
    with container.sync_context():
        test_first, no_cache_first = _injectee()  # type: ignore[call-arg]
        test_second, no_cache_second = _injectee()  # type: ignore[call-arg]

    assert test_first is test_second
    assert test_first is not no_cache_first
    assert test_second is not no_cache_second
    assert no_cache_first is not no_cache_second


@pytest.mark.anyio
async def test_simple_service(container: Container) -> None:
    with container.sync_context() as ctx:
        session = ctx.resolve(_Session)
        assert isinstance(session, _Session)

    async with container.context() as ctx:
        session = await ctx.resolve(_Session)
        assert isinstance(session, _Session)


@pytest.mark.anyio
async def test_retrieve_service_with_dependencies(
    container: Container,
) -> None:
    with container.sync_context() as ctx:
        service = ctx.resolve(_Service)
        assert isinstance(service, _Service)
        assert isinstance(service.session, _Session)

    async with container.context() as ctx:
        service = await ctx.resolve(_Service)
        assert isinstance(service, _Service)
        assert isinstance(service.session, _Session)


@pytest.mark.anyio
async def test_service_with_multiple_dependencies_with_same_type(
    container: Container,
) -> None:
    with container.sync_context() as ctx:
        service = ctx.resolve(_NeedsMultipleImplementations)
        assert isinstance(service.a, _ImplementationA)
        assert isinstance(service.b, _ImplementationB)

    async with container.context() as ctx:
        service = await ctx.resolve(_NeedsMultipleImplementations)
        assert isinstance(service.a, _ImplementationA)
        assert isinstance(service.b, _ImplementationB)


@pytest.mark.anyio
async def test_inject_using_container(
    container: Container,
) -> None:
    @inject(inject_method=InjectMethod.container)
    async def injectee(service: Annotated[_Service, Inject]) -> _Service:
        return service

    token = container_var.set(container)
    # This is fine
    coro = injectee()  # type: ignore[call-arg]
    assert isinstance(await coro, _Service)
    container_var.reset(token)



import pytest

from aioinject import Provider, Singleton


class _Test:
    pass


def test_identity() -> None:
    provider = Singleton(_Test)
    instance = provider.provide_sync()
    assert instance is provider.provide_sync()


@pytest.mark.anyio
async def test_identity_async() -> None:
    provider = Singleton(_Test)
    instance = provider.provide_sync()
    assert instance is provider.provide_sync() is await provider.provide()


@pytest.mark.anyio
async def test_async_function() -> None:
    async def create_test() -> _Test:
        return _Test()

    provider = Singleton[_Test](factory=create_test)
    instance = await provider.provide()
    assert instance is await provider.provide()


@pytest.mark.anyio
async def test_should_not_provide_none_twice() -> None:
    count = 0

    async def func() -> None:
        nonlocal count
        count += 1

    provider: Provider[None] = Singleton(func)
    for _ in range(5):
        assert await provider.provide() is None  # type: ignore[func-returns-value]
        assert count == 1



from collections import defaultdict
from collections.abc import AsyncGenerator, AsyncIterator, Generator, Iterator
from typing import Annotated
from unittest.mock import patch

import pytest

from aioinject import Inject, Provider, providers
from aioinject.providers import Dependency


class _Test:
    pass


@pytest.fixture
def provider() -> Provider[_Test]:
    return providers.Scoped(_Test)


def test_can_provide(provider: Provider[_Test]) -> None:
    instance = provider.provide_sync()
    assert isinstance(instance, _Test)


def test_provided_instances_are_unique(provider: Provider[_Test]) -> None:
    first = provider.provide_sync()
    second = provider.provide_sync()
    assert first is not second


def test_would_pass_kwargs_into_impl(provider: Provider[_Test]) -> None:
    with patch.object(provider, "impl") as factory_mock:
        provider.provide_sync()
        factory_mock.assert_called_once_with()

    kwargs = {"a": 1, "b": 2}
    with patch.object(provider, "impl") as factory_mock:
        provider.provide_sync(**kwargs)
        factory_mock.assert_called_once_with(**kwargs)


def test_would_return_factory_result(provider: Provider[_Test]) -> None:
    instance = object()
    with patch.object(provider, "impl") as factory_mock:
        factory_mock.return_value = instance
        assert provider.provide_sync() is instance


@pytest.mark.anyio
async def test_provide_async() -> None:
    return_value = 42

    async def factory() -> int:
        return return_value

    provider = providers.Scoped[int](factory)
    assert await provider.provide() == return_value


def test_type_hints_on_function() -> None:
    def factory(a: int, b: str) -> None:  # noqa: ARG001
        pass

    provider = providers.Scoped(factory)
    expected = {
        "a": Annotated[int, Inject],
        "b": Annotated[str, Inject],
    }
    assert provider.type_hints == expected


def test_type_hints_on_class() -> None:
    class Test:
        def __init__(self, a: int, b: str) -> None:
            pass

    provider = providers.Scoped(Test)
    expected = {
        "a": Annotated[int, Inject],
        "b": Annotated[str, Inject],
    }
    assert provider.type_hints == expected


def test_annotated_type_hint() -> None:
    def factory(
        a: Annotated[int, Inject(cache=False)],  # noqa: ARG001
    ) -> None:
        pass

    provider = providers.Scoped(factory)
    assert provider.type_hints == {"a": Annotated[int, Inject(cache=False)]}


def test_is_async_on_sync() -> None:
    def factory() -> None:
        pass

    provider = providers.Scoped[None](factory)
    assert not provider.is_async


def test_is_async_on_async() -> None:
    async def factory() -> None:
        pass

    provider = providers.Scoped[None](factory)
    assert provider.is_async


def test_empty_dependencies() -> None:
    def factory() -> None:
        pass

    provider = providers.Scoped(factory)
    assert provider.dependencies == ()


def test_dependencies() -> None:
    def factory(
        a: int,  # noqa: ARG001
        service: Annotated[  # noqa: ARG001
            dict[str, int],
            Inject(defaultdict),
        ],
        string: Annotated[str, Inject(cache=False)],  # noqa: ARG001
    ) -> None:
        pass

    provider = providers.Scoped(factory)
    expected = (
        Dependency(
            name="a",
            type_=int,
            implementation=None,
            use_cache=True,
        ),
        Dependency(
            name="service",
            type_=dict[str, int],
            implementation=defaultdict,
            use_cache=True,
        ),
        Dependency(
            name="string",
            type_=str,
            implementation=None,
            use_cache=False,
        ),
    )
    assert provider.dependencies == expected


def test_generator_return_types() -> None:
    def iterable() -> Iterator[int]:
        yield 42

    def gen() -> Generator[int, None, None]:
        yield 42

    async def async_iterable() -> AsyncIterator[int]:
        yield 42

    async def async_gen() -> AsyncGenerator[int, None]:
        yield 42

    factories = [iterable, gen, async_iterable, async_gen]
    for factory in factories:
        provider = providers.Scoped(factory)
        assert provider.type_ is int



import uuid

import httpx
import pytest


pytestmark = [pytest.mark.anyio]


@pytest.mark.parametrize("resolver_name", ["helloWorld", "helloWorldSync"])
async def test_async_resolver(
    http_client: httpx.AsyncClient,
    provided_value: int,
    resolver_name: str,
) -> None:
    query = """
    query($argument: String!) {
        value: RESOLVER_NAME(argument: $argument)
    }
    """.replace(
        "RESOLVER_NAME",
        resolver_name,
    )
    argument = str(uuid.uuid4())
    response = await http_client.post(
        "",
        json={"query": query, "variables": {"argument": argument}},
    )
    assert response.status_code == httpx.codes.OK.value

    assert response.json() == {
        "data": {"value": f"{argument}-{provided_value}"},
    }


async def test_dataloader(
    http_client: httpx.AsyncClient,
) -> None:
    query = """
    query {
        numbers: dataloader
    }
    """
    str(uuid.uuid4())
    response = await http_client.post("", json={"query": query})
    assert response.status_code == httpx.codes.OK.value

    assert response.json() == {
        "data": {"numbers": list(range(100))},
    }



from aioinject import Container, Object


class _A:
    pass


def test_provider_override() -> None:
    a = _A()
    container = Container()
    container.register(Object(a))
    with container.sync_context() as ctx:
        assert ctx.resolve(_A, use_cache=False) is a

        a_overridden = _A()
        assert a_overridden is not a

        with container.override(Object(a_overridden)):
            assert ctx.resolve(_A, use_cache=False) is a_overridden is not a



from __future__ import annotations

from collections.abc import Callable, Iterator
from typing import TYPE_CHECKING, ParamSpec, TypeVar

from strawberry.extensions import SchemaExtension

from aioinject import decorators, utils
from aioinject.context import container_var


if TYPE_CHECKING:
    from aioinject.containers import Container

_T = TypeVar("_T")
_P = ParamSpec("_P")


def inject(function: Callable[_P, _T]) -> Callable[_P, _T]:
    wrapper = decorators.inject(
        function,
        inject_method=decorators.InjectMethod.container,
    )
    return utils.clear_wrapper(wrapper)


class AioInjectExtension(SchemaExtension):
    def __init__(self, container: Container) -> None:
        self.container = container

    def on_operation(
        self,
    ) -> Iterator[None]:
        token = container_var.set(self.container)
        yield
        container_var.reset(token)



from collections.abc import Sequence
from typing import Any

import pytest

from aioinject import Container, Provider, Scoped
from aioinject.validation import (
    all_dependencies_are_present,
    validate_container,
)
from aioinject.validation.error import (
    ContainerValidationErrorGroup,
    DependencyNotFoundError,
)


_VALIDATORS = [all_dependencies_are_present]


def _str_dependency(number: int) -> str:
    return str(number)


@pytest.mark.parametrize(
    "providers",
    [
        [Scoped(int)],
        [Scoped(_str_dependency), Scoped(int)],
    ],
)
def test_ok(providers: Sequence[Provider[Any]]) -> None:
    container = Container()
    for provider in providers:
        container.register(provider)

    validate_container(container, _VALIDATORS)


def test_err() -> None:
    container = Container()
    container.register(Scoped(_str_dependency))

    with pytest.raises(ContainerValidationErrorGroup) as exc_info:
        validate_container(container, _VALIDATORS)

    assert len(exc_info.value.errors) == 1
    err = exc_info.value.errors[0]
    assert isinstance(err, DependencyNotFoundError)
    assert err.dependency == int
